
import numpy as np
from typing import Union

def Lambda(x:Union[float,int,list,np.ndarray]) -> np.array:
    """DCT scaling function.
    
    :param x: Input value or array.
    :type x: float | list | np.ndarray
    :return: sqrt(1/2) for x=0, otherwise 1
    :rtype: float
    
    :Examples:
    
    Example of call with scalar
    
    >>> Lambda(0) # -> [0.70710678]
    
    Example of call with vector
    
    >>> Lambda([-1,0,1]) # -> [1. 0.70710678 1.]
    """
    y = np.array(x, dtype=np.float64)
    y[y != 0] = 1.
    y[y == 0] = 1/np.sqrt(2)
    return y





from cmath import inf
from distutils.log import info
import logging
import numpy as np
#from pathlib import Path
import tempfile
import typing

from .dct_jpeg import DCTJPEG
from .spatial_jpeg import SpatialJPEG
from . import _jpeg
from ._colorspace import Colorspace
from ._dctmethod import DCTMethod
from ._dithermode import Dithermode

def read_dct(
    path: str
) -> DCTJPEG:
    """Function for reading of the JPEG DCT coefficients.
    
    The file content is loaded once at the call. Then the operations are independent on the source file.
    
    Reading of DCT is a lossless operation, does not perform the JPEG file decompression.
    
    :param path: Path to a source file in JPEG format.
    :type path: str
    :return: DCT JPEG object
    :rtype: :class:`DCTJPEG`
    :raises [IOError]: When source file does not exist

    :Example:
    
    >>> jpeg = jpeglib.read_dct("input.jpeg")
    >>> jpeg.Y; jpeg.Cb; jpeg.Cr; jpeg.qt
    
    In some cases, libjpeg exits the program. We work on improving this and replace exit with "soft" Python exception.
    
    >>> try:
    >>>     jpeg = jpeglib.read_dct("does_not_exist.jpeg")
    >>> except IOError: # raised, file does not exist
    >>>     pass
    
    After call, only the parameters of JPEG are read, so that it is known, how large buffers need to be allocated.
    The allocation and reading of the data happens on the first query.
    
    >>> jpeg = jpeglib.read_dct("input.jpeg")
    >>> 
    >>> # at this point, internally jpeg.Y is None
    >>> # however on first query, it is read
    >>> print(jpeg.Y) # read and returned
    >>> print(jpeg.Y) # second time it is already stored in the object
    >>> print(jpeg.Cb) # no reading, already stored in the object too
    """
    # load file content
    with open(path, "rb") as f:
        content = f.read()
    # load info
    info = _jpeg.load_jpeg_info(path)
    # create jpeg
    return DCTJPEG(
        path=path,
        content=content,
        height=info.height,
        width=info.width,
        block_dims=info.block_dims,
        samp_factor=info.samp_factor,
        jpeg_color_space=info.jpeg_color_space,
        #num_components=info.num_components,
        quant_tbl_no=None,
        markers=info.markers,
        Y=None,
        Cb=None,
        Cr=None,
        qt=None,
        progressive_mode=info.progressive_mode
    )


def read_spatial(
    path: str,
    out_color_space: str = None,
    dither_mode: str = None,
    dct_method: str = None,
    flags:list = []
) -> SpatialJPEG:
    """Function for decompressing the JPEG as a pixel data (spatial domain).
    
    The file content is loaded once at the call. Then the operations are independent on the source file.
    
    In some cases, libjpeg exits the program. We work on improving this and replace exit with "soft" Python exception.
    
    :param path: Path to a source file in JPEG format.
    :type path: str
    :param out_color_space: Output color space, must be accepted by :class:`_colorspace.Colorspace`. If not given, using the libjpeg default.
    :type out_color_space: str, optional
    :param dither_mode: Dither mode, must be accepted by :class:`_dithermode.Dithermode`. If not given, using the libjpeg default.
    :type dither_mode: str, optional
    :param dct_method: DCT method, must be accepted by :class:`_dctmethod.DCTMethod`. If not given, using the libjpeg default.
    :type dct_method: str, optional
    :param flags: Bool decompression parameters as str. If not given, using the libjpeg default. Read more at `glossary <https://jpeglib.readthedocs.io/en/latest/glossary.html#flags>`_.
    :type flags: list, optional
    :return: Spatial JPEG object
    :rtype: :class:`SpatialJPEG`
    :raises [IOError]: When source file does not exist
    
    :Example:
    
    >>> im = jpeglib.read_spatial("input.jpeg")
    >>> im.spatial
    
    >>> try:
    >>>     im = jpeglib.read_spatial("does_not_exist.jpeg")
    >>> except IOError: # raised, file does not exist
    >>>     pass
    
    After call, only the parameters of JPEG are read, so that it is known, how large buffer needs to be allocated.
    The allocation and reading of the data happens on the first query.
    
    >>> im = jpeglib.read_spatial("input.jpeg")
    >>> # at this point, internally jpeg.spatial is None
    >>> # however on first query, it is read
    >>> print(im.spatial) # read and returned
    >>> print(im.spatial) # second time it is already stored in the object 
    """
    # load file content
    with open(path, "rb") as f:
        content = f.read()
    # load info
    info = _jpeg.load_jpeg_info(path)
    # parse
    num_components = info.jpeg_color_space
    if out_color_space is not None:
        out_color_space = Colorspace(out_color_space)
        num_components = out_color_space.channels
    if dct_method is not None:
        dct_method = DCTMethod(dct_method)
    if dither_mode is not None:
        dither_mode = Dithermode(dither_mode)
    # create jpeg
    return SpatialJPEG(
        path                = path,
        content             = content,
        height              = info.height,
        width               = info.width,
        block_dims          = info.block_dims,
        samp_factor         = info.samp_factor,
        jpeg_color_space    = info.jpeg_color_space,
        #num_components      = num_components,
        markers             = info.markers,
        spatial             = None,
        color_space         = out_color_space,
        dither_mode         = dither_mode,
        dct_method          = dct_method,
        flags               = flags,
        progressive_mode    = info.progressive_mode
    )

def from_spatial(
    spatial: np.ndarray,
    in_color_space:typing.Union[str,Colorspace] = None
) -> SpatialJPEG:
    """A factory of :class:`SpatialJPEG` from pixel data.
    
    The color space inference is based on number of color channels.
    For single channel, grayscale is assumed. For three channels, rgb is assumed.
    
    .. warning::
        Parameter :obj:`SpatialJPEG.path` is not initialized.
        When calling :meth:`SpatialJPEG.write_spatial`, you have to specify `path`,
        otherwise an error is raised.
        
    :param spatial: Spatial representation.
    :type spatial: np.ndarray
    :param in_color_space: Color space of the input. If not given, infered from the shape.
    :type in_color_space: str | Colorspace, optional
    :raises [IOError]: When color space can't be infered.
    
    :Example:
    
    When data has three color channels (in the dimension 2), rgb is infered.
    
    >>> spatial = np.random.randint(0,255,(16,16,3),dtype=np.uint8)
    >>> im = jpeglib.from_spatial(spatial) # 3 channels -> rgb infered
    >>> print(im.color_space) # -> 'JCS_RGB'
    >>> im.write_spatial("output.jpeg")
    
    When data has one color channels, grayscale is infered
    
    >>> spatial = np.random.randint(0,255,(16,16,1),dtype=np.uint8)
    >>> im = jpeglib.from_spatial(spatial) # 1 channel -> grayscale infered
    >>> print(im.color_space) # -> 'JCS_GRAYSCALE'
    >>> im.write_spatial("output.jpeg")
    
    For other color channels, color space can't be infered. Error is raised.
    
    >>> spatial = np.random.randint(0,255,(16,16,7),dtype=np.uint8)
    >>> try:
    >>>     im = jpeglib.from_spatial(spatial)
    >>> except IOError:
    >>>     raised
    
    When output is not specified when writing, error is raised.
    
    >>> spatial = np.random.randint(0,255,(16,16,3),dtype=np.uint8)
    >>> im = jpeglib.from_spatial(spatial)
    >>> try:
    >>>     im.write_spatial()
    >>> except IOError:
    >>>     pass
    """
    # shape
    height, width, num_components = spatial.shape
    # parse colorspace
    if in_color_space is not None:
        try:
            in_color_space = Colorspace(in_color_space)
        except:
            in_color_space = in_color_space
    # infere colorspace
    if in_color_space is None:
        if num_components == 3:
            in_color_space = Colorspace('JCS_RGB')
            jpeg_color_space = Colorspace('JCS_YCbCr')
        elif num_components == 1:
            in_color_space = Colorspace('JCS_GRAYSCALE')
            jpeg_color_space = Colorspace('JCS_GRAYSCALE')
        else:
            raise IOError('failed to infere colorspace')
    # create jpeg
    return SpatialJPEG(
        path=None,
        content=None,
        height=height,
        width=width,
        block_dims=None,
        samp_factor=None,
        jpeg_color_space=jpeg_color_space,
        markers=None,
        spatial=spatial,
        color_space=in_color_space,
        dither_mode=None,
        dct_method=None,
        flags=[],
        progressive_mode=False
    )

def from_dct(
    Y: np.ndarray,
    Cb: np.ndarray = None,
    Cr: np.ndarray = None,
    qt: np.ndarray = None,
    quant_tbl_no: list = None,
) -> DCTJPEG:
    """A factory of :class:`DCTJPEG` from DCT data.
    
    The color space inference is based on number of color channels.
    If chrominance is not given, grayscale JPEG is assumed. For three channels, YCbCr JPEG is assumed.

    The quantization table assigment is infered based on components and number of quantization tables given.
    If you wish a custom assignment, use :obj:`DCTJPEG.qt`.

    .. warning::
        Parameter :obj:`DCTJPEG.path` is not initialized.
        When calling :meth:`DCTJPEG.write_dct`, you have to specify `path`,
        otherwise an error is raised.
        
    :param Y: Luminance tensor.
    :type Y: np.ndarray
    :param Cb: Blue chrominance tensor.
    :type Cb: np.ndarray
    :param Cr: Red chrominance tensor.
    :type Cr: np.ndarray
    :param qt: Quantization table tensor.
    :type qt: np.ndarray
    
    :Example:
    
    >>> Y = np.random.randint(-127, 127,(2,2,8,8),dtype=np.int16)
    >>> Cb = np.random.randint(-127, 127,(1,1,8,8),dtype=np.int16)
    >>> Cr = np.random.randint(-127, 127,(1,1,8,8),dtype=np.int16)
    >>> im = jpeglib.from_dct(Y, Cb, Cr, qt=75) # chrominance -> YCbCr infered

    When data has one color channels, grayscale is infered
    
    >>> Y = np.random.randint(-127, 127,(2,2,8,8),dtype=np.int16)
    >>> im = jpeglib.from_dct(Y) # chrominance -> YCbCr infered
    
    For other color channels, color space can't be infered. Error is raised.
    
    >>> Y = np.random.randint(-127, 127,(2,2,8,8),dtype=np.int16)
    >>> X = np.random.randint(-127, 127,(1,1,8,8),dtype=np.int16)
    >>> try:
    >>>     im = jpeglib.from_dct(Y, X)
    >>> except IOError:
    >>>     raised
    
    When output is not specified when writing, error is raised.
    
    >>> Y = np.random.randint(-127, 127,(2,2,8,8),dtype=np.int16)
    >>> im = jpeglib.from_dct(Y)
    >>> try:
    >>>     im.write_dct()
    >>> except IOError:
    >>>     pass
    """
    # infere colorspace
    if Cb is not None and Cr is not None:
        jpeg_color_space = Colorspace('JCS_YCbCr')
    elif Cb is None and Cr is None:
        jpeg_color_space = Colorspace('JCS_YCbCr') # only Y
        #jpeg_color_space = Colorspace('JCS_GRAYSCALE')
    else:
        raise IOError('failed to infere colorspace')
    # infere quant_tbl_no
    if quant_tbl_no is None and qt is not None:
        quant_tbl_no = np.array([0,0,0,0])
        if Cb is not None and Cr is not None:
            if qt.shape[0] == 2:
                quant_tbl_no = np.array([0,1,1,0])
            elif qt.shape[0] == 3:
                quant_tbl_no = np.array([0,1,2,0])
    # infere samp_factor
    # TODO
    samp_factor = ((1,1),(1,1),(1,1))
    # block dims
    block_dims = [[Y.shape[0],Y.shape[1]]]
    if Cb is not None and Cr is not None:
        block_dims.append([Cb.shape[0],Cb.shape[1]])
        block_dims.append([Cr.shape[0],Cr.shape[1]])
    block_dims = np.array(block_dims)

    # create jpeg
    return DCTJPEG(
        path=None,
        content=None,
        height=Y.shape[0]*Y.shape[2],
        width=Y.shape[1]*Y.shape[3],
        block_dims=block_dims,
        samp_factor=samp_factor,
        jpeg_color_space=jpeg_color_space,
        #num_components=info.num_components,
        quant_tbl_no=quant_tbl_no,
        markers=None,
        Y=Y,
        Cb=Cb,
        Cr=Cr,
        qt=qt,
        progressive_mode=None,
    )



    # shape
    height, width, num_components = spatial.shape
    # parse colorspace
    if in_color_space is not None:
        try:
            in_color_space = Colorspace(in_color_space)
        except:
            in_color_space = in_color_space
    # infere colorspace
    if in_color_space is None:
        if num_components == 3:
            in_color_space = Colorspace('JCS_RGB')
        elif num_components == 1:
            in_color_space = Colorspace('JCS_GRAYSCALE')
        else:
            raise IOError('failed to infere colorspace')
    # create jpeg
    return SpatialJPEG(
        path=None,
        content=None,
        height=height,
        width=width,
        block_dims=None,
        samp_factor=None,
        jpeg_color_space=None,
        markers=None,
        spatial=spatial,
        color_space=in_color_space,
        dither_mode=None,
        dct_method=None,
        flags=[],
        progressive_mode=False
    )

    # with JPEG(srcfile) as im:
    #    Y,CbCr,qt = im.read_dct(*args, **kwargs)

# def write_dct(Y, CbCr=None, dstfile=None, qt=None, *args, **kwargs):
#     """Writes DCT coefficients to a file.

#     :param srcfile: Path to a source file in JPEG format. An implicit destination for writing.
#     :type srcfile: str
#     :param args: Positional arguments passed to JPEG.write_dct.
#     :param kwargs: Named arguments passed to JPEG.write_dct.

#     More information in documentation of JPEG.write_dct.

#     :Example:

#     >>> Y,CbCr,qt = jpeglib.read_dct("input.jpeg")
#     >>> jpeglib.write_dct(Y, CbCr, "output.jpeg", qt=qt)
#     """
#     assert(dstfile is not None)
#     with JPEG() as im:
#         im.write_dct(Y, CbCr, dstfile=dstfile, qt=qt, *args, **kwargs)

# def read_spatial(srcfile, *args, **kwargs):
#     """Decompresses the file into the spatial domain.

#     :param srcfile: Path to a source file in JPEG format. An implicit destination for writing.
#     :type srcfile: str
#     :param args: Positional arguments passed to JPEG.read_spatial.
#     :param kwargs: Named arguments passed to JPEG.read_spatial.

#     More information in documentation of JPEG.read_spatial.

#     :Example:

#     >>> x = jpeglib.read_spatial("input.jpeg")
#     >>> import matplotlib.pyplot as plt; plt.imshow(x)
#     """
#     with JPEG(srcfile) as im:
#         x = im.read_spatial(*args, **kwargs)
#         return x

# def write_spatial(data, dstfile, *args, **kwargs):
#     """Writes spatial image representation (i.e. RGB) to a file.

#     :param data: Numpy array with spatial data.
#     :type data: numpy.ndarray
#     :param dstfile: Destination file name. If not given, source file is overwritten.
#     :type dstfile: str, optional
#     :param args: Positional arguments passed to JPEG.write_spatial.
#     :param kwargs: Named arguments passed to JPEG.write_spatial.

#     More information in documentation of JPEG.read_spatial.

#     :Example:

#     >>> x = jpeglib.read_spatial("input.jpeg")
#     >>> jpeglib.write_spatial(x, "output.jpeg")
#     """
#     with JPEG() as im:
#         im.write_spatial(data, dstfile, *args, **kwargs)




import ctypes
from dataclasses import dataclass,fields
import numpy as np
import tempfile
from ._bind import CJpegLib
from . import _jpeg

@dataclass
class DCTJPEG(_jpeg.JPEG):
    """JPEG instance to work with DCT domain."""
    
    Y: np.ndarray
    """luminance tensor"""
    Cb: np.ndarray
    """chrominance blue-difference tensor"""
    Cr: np.ndarray
    """chrominance red-difference tensor"""
    qt: np.ndarray
    """quantization tensor"""
    quant_tbl_no: np.ndarray
    """assignment of quantization tables to components, (0 Y, 1 Cb, 1Cr) by default"""
    
    # def is_read(self) -> bool:
    #     has_Y = self.Y is not None
    #     has_qt = self.qt is not None
    #     has_CbCr = self.Cb is not None and self.Cr is not None
    #     has_no_CbCr = self.Cb is None and self.Cr is None
    #     return (
    #         (has_Y and has_qt and has_no_CbCr) # grayscale
    #         or (has_Y and has_qt and has_CbCr) # color
    #     )
    
    def _alloc_dct_component(self, i:int):
        return (((ctypes.c_short * 64) * self.width_in_blocks(i)) * self.height_in_blocks(i))()
    
    def read_dct(self) -> tuple:
        """Function to allocate the buffer and read the image data.
        
        :return: luminance Y and chrominance Cb, Cr tensors and quantization table qt as Y,(Cb,Cr),qt
        :rtype: tuple
        
        :Example:
        
        >>> jpeg = jpeglib.read_dct("input.jpeg")
        >>> Y,(Cb,Cr),qt = jpeg.read_dct()
        
        Function is internally called when
        
        >>> jpeg = jpeglib,read_dct("input.jpeg")
        >>> jpeg.Y # jpeg.read_dct() called internally
        """
        # no content
        if self.content is None:
            return None,(None,None),None
        # write content into temporary file
        tmp = tempfile.NamedTemporaryFile(suffix='.jpeg')
        tmp.write(self.content)
        tmp.flush()
        # allocate DCT components
        Y = self._alloc_dct_component(0)
        Cb,Cr = None,None
        if self.has_chrominance: # has chrominance
            Cb = self._alloc_dct_component(1)
            Cr = self._alloc_dct_component(2)
        qt = ((ctypes.c_short * 64) * 4)()
        _quant_tbl_no = (ctypes.c_ubyte*4)()
        # call
        CJpegLib.read_jpeg_dct(
            path            = self.path,
            srcfile         = tmp.name,
            Y               = Y,
            Cb              = Cb,
            Cr              = Cr,
            qt              = qt,
            quant_tbl_no    =_quant_tbl_no,
        )
        # close temporary file
        tmp.close()
        # process
        def process_component(comp):
            comp_np = np.ctypeslib.as_array(comp)
            return comp_np.reshape((*comp_np.shape[:-1],8,8))
        qt = process_component(qt)
        self.Y = process_component(Y)
        if self.has_chrominance:
            self.Cb = process_component(Cb)
            self.Cr = process_component(Cr)
        self.quant_tbl_no = np.array([_quant_tbl_no[i] for i in range(self.num_components)])
        # crop
        self.qt = qt[:self.num_components]
        # return
        return self.Y,(self.Cb,self.Cr),self.qt
    
    def write_dct(self, path:str = None, quality:int=-1):
        """Function to write DCT coefficients to JPEG file.
        
        Does not perform JPEG compression, writing DCT is lossless.
        
        :param path: Destination file name. If not given, source file is overwritten.
        :type path: str, optional
        :param quality: Compression quality, between 0 and 100. Special value -1 stands for using qt inside the instance or keeping libjpeg default.
        :type quality: int, optional
        
        :Example:
        
        >>> jpeg = jpeglib.read_spatial("input.jpeg")
        >>> jpeg.write_spatial("output.jpeg", quality=92)
        """
        tmp = tempfile.NamedTemporaryFile(suffix='jpeg')
        if self.content is not None:
            # write content into temporary file
            tmp.write(self.content)
            tmp.flush()
        # path
        dstfile = path if path is not None else self.path
        if dstfile is None:
            raise IOError('no destination file specified')
        # convert dct
        def process_component(comp):
            if comp is None:
                return None
            comp = comp.reshape((*comp.shape[:-2],64))
            return np.ctypeslib.as_ctypes(comp.astype(np.int16))
        qt = process_component(self.qt)
        Y = process_component(self.Y)
        Cb = process_component(self.Cb)
        Cr = process_component(self.Cr)
        # quality and quantization table
        assert(quality in set(range(-1,101)))
        if quality != -1:
            qt = None
            
        # call
        CJpegLib.write_jpeg_dct(
            srcfile         = tmp.name if self.content is not None else None,
            dstfile         = dstfile,
            Y               = Y,
            Cb              = Cb,
            Cr              = Cr,
            image_dims      = self.c_image_dims(),
            block_dims      = self.c_block_dims(),
            in_color_space  = self.jpeg_color_space.index,
            in_components   = self.num_components,
            qt              = qt,
            quality         = quality,
            num_markers     = self.num_markers,
            marker_types    = self.c_marker_types(),
            marker_lengths  = self.c_marker_lengths(),
            markers         = self.c_markers(),
        )
    
    @property
    def Y(self) -> np.ndarray:
        """Luminance tensor getter."""
        if self._Y is None:
            self.read_dct()
        return self._Y
    @Y.setter
    def Y(self, Y: np.ndarray):
        """Luminance tensor setter."""
        self._Y = Y
    
    @property
    def Cb(self) -> np.ndarray:
        """Chrominance blue-difference tensor getter."""
        if self.has_chrominance and self._Cb is None:
            self.read_dct()
        return self._Cb
    @Cb.setter
    def Cb(self, Cb: np.ndarray):
        """Chrominance blue-difference tensor setter."""
        self._Cb = Cb
    
    @property
    def Cr(self) -> np.ndarray:
        """Chrominance red-difference tensor getter."""
        if self.has_chrominance and self._Cr is None:
            self.read_dct()
        return self._Cr
    @Cr.setter
    def Cr(self, Cr: np.ndarray):
        """Chrominance red-difference tensor setter."""
        self._Cr = Cr
    
    @property
    def qt(self) -> np.ndarray:
        """Quantization table getter."""
        if self._qt is None:
            self.read_dct()
        return self._qt
    @qt.setter
    def qt(self, qt: np.ndarray):
        """Quantization table setter."""
        self._qt = qt
    @property
    def quant_tbl_no(self) -> list:
        """Getter of assignment of quantization tables to components"""
        if self._quant_tbl_no is None:
            self.read_dct()
        return self._quant_tbl_no
    @quant_tbl_no.setter
    def quant_tbl_no(self, quant_tbl_no:list):
        """Setter of assignment of quantization tables to components"""
        self._quant_tbl_no = quant_tbl_no
    
    def get_component_qt(self, idx:int) -> np.ndarray:
        """Getter of the quantization table of a component, dependent on assignment.
        
        :param idx: Component index.
        :type idx: int
        :return: quantization table of the component
        :rtype: np.ndarray
        
        :Examples:
        
        To retreive a component quantization table, simply type
        
        >>> jpeg = jpeglib.read_dct("input.jpeg")
        >>> # retrieval of component QTs
        >>> qtY = jpeg.get_component_qt(0)
        >>> qtCb = jpeg.get_component_qt(1)
        >>> qtCr = jpeg.get_component_qt(2)
        
        Usual assignment is Y having qt[0] and Cb and Cr sharing qt[1]. However it may differ.
        
        >>> # retrieval of QTs at index
        >>> qt0 = jpeg.qt[0] # usually qtY
        >>> qt1 = jpeg.qt[1] # usually qtCb, qtCr (shared)
        
        E.g. ffmpeg uses a single quantization matrix for all three channels.
        """
        return self.qt[self.quant_tbl_no[idx]]
    
    def free(self):
        """Free the allocated tensors."""
        del self._Y
        del self._Cb
        del self._Cr
        del self._qt

        
@dataclass
class DCTJPEGio(DCTJPEG):
    """Class for compatiblity with jpegio."""
    
    coef_arrays: list
    """DCT coefficient arrays in jpegio format"""
    quant_tables: list
    """quantization tables in jpegio format"""
    
    def __post_init__(self):
        self._jpeg_to_jpegio()
    def _jpeg_to_jpegio(self):
        # colored image
        if self.has_chrominance:
            self._coef_arrays = [
                self._convert_dct_jpegio(self.Y),
                self._convert_dct_jpegio(self.Cb),
                self._convert_dct_jpegio(self.Cr)
            ]
            self._quant_tables = [
                self.get_component_qt(0).astype(np.int32),
                self.get_component_qt(1).astype(np.int32)
            ]
        # grayscale
        else:
            self._coef_arrays = [self._convert_dct_jpegio(self.Y)]
            self._quant_tables = [self.get_component_qt(0).astype(np.int32)]

            
    def _jpegio_to_jpeg(self):
        self.Y = self._convert_jpegio_dct(self.coef_arrays[0])
        if self.has_chrominance:
            self.Cb = self._convert_jpegio_dct(self.coef_arrays[1])
            self.Cr = self._convert_jpegio_dct(self.coef_arrays[2])
        if self.has_chrominance:
            self.qt = np.array([
                self.quant_tables[0],
                self.quant_tables[1]
            ], dtype=np.uint16)
            self.quant_tbl_no = np.array([0,1,1])
        else:
            self.qt = np.array([
                self.quant_tables[0]
            ], dtype=np.uint16)
            self.quant_tbl_no = np.array([0])
    
    def _convert_dct_jpegio(self, dct:np.ndarray):
        return (dct
            .transpose((0,3,1,2))
            .reshape((dct.shape[0]*dct.shape[2], dct.shape[1]*dct.shape[3]))
        ).astype(np.int32)
    def _convert_jpegio_dct(self, dct:np.ndarray):
        return (dct
            .reshape((-1,8,dct.shape[1]//8,8))
            .transpose((0,2,3,1))
        ).astype(np.int16)
    
    @property
    def coef_arrays(self) -> list:
        """Convertor of DCT coefficients to jpegio format."""
        return self._coef_arrays

    @coef_arrays.setter
    def coef_arrays(self, coef_arrays:list):
        """Setter of coefficient arrays in jpegio format."""
        self._coef_arrays = coef_arrays
    
    @property
    def quant_tables(self) -> list:
        """Convertor of quantization tables to jpegio format."""
        return self._quant_tables
    @quant_tables.setter
    def quant_tables(self, quant_tables:list):
        """Setter of quantization tables in jpegio format."""
        self._quant_tables = quant_tables
    
    def write(self, fpath:str, flags:int=-1, quality:int=-1):
        """Function to write DCT coefficients in jpegio format to JPEG file.
    
        Does not perform JPEG compression, writing DCT is lossless.
    
        :param obj:
        :type obj: str
        :param fpath: Destination file name. If not given, source file is overwritten.
        :type fpath: str, optional
        :param flags: Flags. For backward compatibility, does not make a difference.
        :type flags: int
        :param quality: Compression quality, between 0 and 100. Special value -1 stands for using qt inside the instance or keeping libjpeg default.
        :type quality: int, optional
    
        :Example:
    
        >>> jpeg = jpeglib.read_dct("input.jpeg")
        >>> jpeg = jpeglib.to_jpegio(jpeg)
        >>> jpeg.write("output.jpeg", quality=92)
        """
        # pass data from jpegio to jpeg
        self._jpegio_to_jpeg()
        # write 
        self.write_dct(path=fpath, quality=quality)
        
def to_jpegio(jpeg: DCTJPEG):
    """Convertor of object of :class:`DCTJPEG` to :class:`DCTJPEGio`.
    
    When :class:`DCTJPEG` is converted to :class:`DCTJPEGio`,
    the data objects :class:`DCTJPEG` are not updated until writing.
    This behavior will be changed in the future.
    
    :param jpeg: JPEG object to convert
    :type jpeg: :class:`DCTJPEG`
    :return: converted JPEG object
    :rtype: :class:`DCTJPEGio`
    
    :Example:
    
    >>> jpeg = jpeglib.read_dct("input.jpeg")
    >>> jpeg = jpeglib.to_jpegio(jpeg)
    >>> jpeg.coef_arrays; jpeg.quant_tables
    """
    vals = {field.name: getattr(jpeg,field.name) for field in fields(jpeg)}
    return DCTJPEGio(**vals)




from math import cos,pi
import numpy as np

from ._base import *

def _dct_ii(X:np.ndarray, u:int) -> float:
    """DCT coefficient at index u.
    
    :param X: input pixel data tensor 
    :type X: np.ndarray
    :param u: index of DCT coefficient
    :type u: int
    :return: output DCT coefficient value
    :rtype: float
    
    :Examples:
    
    >>> _dct_ii([1,2,1], 1) # -> 2.72e-16
    """
    N = len(x)
    Yu = np.sqrt(2/N) * Lambda(u) * sum(X[i] * np.cos(np.pi/N*(i+.5)*u) for i in range(N))
    return Yu

def _dct_iii(Y:np.ndarray, i:int) -> float:
    """Pixel value at index i.
    
    :param x: input DCT coefficient tensor
    :type x: np.ndarray
    :param i: index of the pixel
    :type i: int
    :return: output pixel value
    :rtype: float
    
    :Examples:
    
    >>> _dct_iii([2,1,-1], 1) # -> 1.97
    """
    N = len(x)
    Xi = np.sqrt(2/N) * sum(Lambda(u) * Y[u] * np.cos(np.pi/N*(i+.5)*u) for u in range(N))
    return Xi

def DCT1D(X:np.ndarray) -> np.ndarray:
    """Python implementation of 1D DCT transformation.
    
    :param X: pixel data in spatial domain
    :type X: np.ndarray
    :return: tensor of DCT coefficients
    :rtype: np.ndarray
    
    :Examples:
    
    >>> X = np.array([1,2,1])
    >>> Y = jpeglib.dct.DCT1D(X)
    """
    return [_dct_ii(X,k) for k in range(len(X))]

def iDCT1D(Y:np.ndarray):
    """Python implementation of 1D inverse DCT transformation.
    
    :param Y: tensor of DCT coefficients
    :type Y: np.ndarray
    :return: pixel data in spatial domain
    :rtype: np.ndarray
    
    :Examples:
    
    >>> X = np.array([1,2,1])
    >>> Y = jpeglib.dct.DCT1D(X)
    >>> jpeglib.dct.iDCT1D(Y) # -> reconstruction of X
    """
    return [_dct_iii(Y,k) for k in range(len(Y))]

__all__ = ["DCT1D", "iDCT1D"]



import ctypes
from collections import namedtuple
from dataclasses import dataclass
import logging
import numpy as np
from sys import flags

from ._bind import CJpegLib
from ._colorspace import Colorspace
from ._marker import Marker


@dataclass
class JPEG:
    """JPEG abstract class."""
    
    path: str
    """path to the jpeg file"""
    content: bytes
    """cached binary content of the jpeg"""
    height: int
    """image height"""
    width: int
    """image width"""
    block_dims: np.ndarray
    """DCT shapes in blocks; first is DCT component (0 Y,1 Cb,2 Cr), second is dimension (0 height, 1 width)"""
    samp_factor: np.ndarray
    """sampling factor; first is DCT component, (0 Y,1 Cb,2 Cr), second is orientation (0 horizontal, 1 vertical)"""
    jpeg_color_space: Colorspace
    """color space of the JPEG file"""
    markers: list
    """list of marker objects"""
    progressive_mode: bool
    """indicator of progressive (True) or sequential (False) JPEG"""
    
    def height_in_blocks(self, component:int) -> int:
        """Getter of height in blocks.
        
        :param component: chroma component index (0 Y, 1 Cb, 2 Cr)
        :type component: int
        :return: chroma component height
        :rtype: int
        :raises [IndexError]: when component index is out of range, dependent on number of components
        
        :Example:
        
        >>> im = jpeglib.read_spatial("input.jpeg")
        >>> im.height_in_blocks(0) # -> math.ceil(im.height/8)
        
        Block dimension takes into account the chroma sampling factor for chrominance channels.
        
        >>> im = jpeglib.read_spatial("input.jpeg")
        >>> im.height_in_blocks(1) # -> math.ceil(im.height/8 * im.samp_factor[1][0]/im.samp_factor[0][0])
        >>> im.height_in_blocks(2) # -> math.ceil(im.height/8 * im.samp_factor[2][0]/im.samp_factor[0][0])
        
        Block dimensions are not initialized, when constructing from spatial domain.
        
        >>> spatial = np.random.randint(0,255,(32,32,1),dtype=np.uint8)
        >>> im = jpeglib.from_spatial(spatial)
        >>> im.height_in_blocks(0) # -> None
        """
        if self.block_dims is None:
            return None
        return self.block_dims[component][0]

    def width_in_blocks(self, component:int) -> int:
        """Getter of width in blocks.
        
        :param component: chroma component index (0 Y, 1 Cb, 2 Cr)
        :type component: int
        :return: chroma component width
        :rtype: int
        :raises [IndexError]: when component index is out of range, dependent on number of components
        
        :Example:
        
        >>> im = jpeglib.read_spatial("input.jpeg")
        >>> im.width_in_blocks(0) # -> math.ceil(im.width/8)
        
        Block dimension takes into account the chroma sampling factor for chrominance channels.
        
        >>> im = jpeglib.read_spatial("input.jpeg")
        >>> im.width_in_blocks(1) # -> math.ceil(im.width/8 * im.samp_factor[1][1]/im.samp_factor[0][1])
        >>> im.width_in_blocks(2) # -> math.ceil(im.width/8 * im.samp_factor[2][1]/im.samp_factor[0][1])
        
        Block dimensions are not initialized, when constructing from spatial domain.
        
        >>> spatial = np.random.randint(0,255,(32,32,1),dtype=np.uint8)
        >>> im = jpeglib.from_spatial(spatial)
        >>> im.height_in_blocks(0) # -> None
        """
        if self.block_dims is None:
            return None
        return self.block_dims[component][1]

    @property
    def has_chrominance(self) -> bool:
        """Indicator of presence of color channels.
        
        :return: True for color image, False for grayscale image
        :rtype: bool
        
        :Example:
        
        >>> im = jpeglib.read_spatial("input.jpeg")
        >>> im.has_chrominance # -> True
        
        >>> im = jpeglib.read_spatial("input.jpeg", "JCS_GRAYSCALE")
        >>> im.has_chrominance # -> False
        
        """
        return self.num_components > 1

    @property
    def num_components(self) -> int:
        """Getter of number of color components in the JPEG.
        
        :return: Number of color components.
        :rtype: bool
        
        :Example:
        
        >>> im = jpeglib.read_spatial("input.jpeg")
        >>> im.has_chrominance # -> 3
        
        >>> im = jpeglib.read_spatial("input.jpeg", "JCS_GRAYSCALE")
        >>> im.has_chrominance # -> 1
        """
        return self.jpeg_color_space.channels
    
    @property
    def num_markers(self) -> int:
        """Getter of number of markers.
        
        :return: Number of markers.
        :rtype: int
        
        :Example:
        
        >>> im = jpeglib.read_spatial("input_2markers.jpeg")
        >>> im.num_markers # -> 2
        
        Synthetic JPEG has no markers in it.
        
        >>> spatial = np.random.randint(0,255,(16,16,3),dtype=np.uint8)
        >>> im = jpeglib.from_spatial(spatial)
        >>> im.num_markers # -> 0
        """
        if self.markers is None:
            return 0
        return len(self.markers)

    def c_image_dims(self):
        return (ctypes.c_int * 2)(self.height, self.width)
    def c_block_dims(self):
        if self.has_chrominance:
            return (ctypes.c_int * 6)(
                self.block_dims[0][0], self.block_dims[0][1],
                self.block_dims[1][0], self.block_dims[1][1],
                self.block_dims[2][0], self.block_dims[2][1],)
        else:
            return (ctypes.c_int * 6)(
                self.block_dims[0][0], self.block_dims[0][1],)
    def c_samp_factor(self):
        if self.samp_factor is None:
            return self.samp_factor
        samp_factor = np.array(self.samp_factor, dtype=np.int32)
        return np.ctypeslib.as_ctypes(samp_factor)
    def c_marker_types(self):
        if self.markers is None:
            return None
        marker_types = [marker.index for marker in self.markers]
        return (ctypes.c_int32*self.num_markers)(*marker_types)
    def c_marker_lengths(self):
        if self.markers is None:
            return None
        marker_lengths = [marker.length for marker in self.markers]
        return (ctypes.c_int32*self.num_markers)(*marker_lengths)
    def c_markers(self):
        if self.markers is None:
            return None
        marker_lengths = [marker.length for marker in self.markers]
        marker_contents = []
        for marker in self.markers:
            marker_contents += [i for i in marker.content]
        return (ctypes.c_ubyte*int(np.sum(marker_lengths)))(*marker_contents)

    def free(self):
        """Free the allocated tensors."""
        raise NotImplementedError
    def close(self):
        """Closes the object. Defined for interface compatibility with PIL.
        
        :Example:

        >>> im = jpeglib.read_dct("input.jpeg")
        >>> # work with im
        >>> im.close()
        """
        pass
    
    def __enter__(self):
        """Method for using ``with`` statement together with :class:`JPEG`.
        
        :Example:
        
        >>> with jpeglib.read_dct("input.jpeg") as im:
        >>>     im.Y; im.Cb; im.Cr; im.qt
        """
        return self
    def __exit__(self, exception_type, exception_val, trace):
        """Method for using ``with`` statement together with :class:`JPEG`."""
        self.close()
        
    


def load_jpeg_info(path: str) -> JPEG:
    """"""
    # allocate
    _block_dims = (ctypes.c_int*6)()
    _image_dims = (ctypes.c_int*2)()
    _num_components = (ctypes.c_int*1)()
    _samp_factor = (ctypes.c_int*6)()
    _jpeg_color_space = (ctypes.c_int*2)()
    
    _marker_lengths = (ctypes.c_int*20)()
    _marker_types = (ctypes.c_uint32*20)()
    _flags = (ctypes.c_uint64*1)()

    # call
    CJpegLib.read_jpeg_info(
        srcfile=str(path),
        block_dims=_block_dims,
        image_dims=_image_dims,
        num_components=_num_components,
        samp_factor=_samp_factor,
        jpeg_color_space=_jpeg_color_space,
        marker_lengths=_marker_lengths,
        marker_types=_marker_types,
        flags=_flags
    )
    # process
    num_components = _num_components[0]  # number of components in JPEG
    block_dims = (
        np.array([_block_dims[i] for i in range(2*num_components)], int)
        .reshape(num_components, 2))
    samp_factor = (
        np.array([_samp_factor[i] for i in range(2*num_components)], int)
        .reshape(num_components, 2))
    
    markers = []
    for i in range(20):
        # marker length
        marker_length = _marker_lengths[i]
        if marker_length == 0:
            break
        # marker type
        marker_type = _marker_types[i]
        # create marker
        marker = Marker.from_index(
            index=marker_type,
            length=marker_length,
            content=None,
        )
        markers.append(marker)
    marker_lengths = np.array(
        [marker.length for marker in markers], dtype=np.int32)
    num_markers = marker_lengths.shape[0]
    flags = CJpegLib.mask_to_flags(_flags)

    # allocate
    _markers = (ctypes.c_ubyte * np.sum(marker_lengths))()
    # call
    CJpegLib.read_jpeg_markers(
        srcfile=str(path),
        markers=_markers,
    )
    # process
    cumlens = np.cumsum([0] + marker_lengths.tolist())
    for i in range(num_markers):
        markers[i].content = bytes(_markers[cumlens[i]:cumlens[i+1]])
    # create jpeg
    return JPEG(
        path=path,
        height=_image_dims[0],
        width=_image_dims[1],
        block_dims=block_dims,
        samp_factor=samp_factor,
        #num_components=num_components,
        #out_color_space=Colorspace.from_index(_jpeg_color_space[0]), # out_color_space
        jpeg_color_space=Colorspace.from_index(_jpeg_color_space[1]),
        content=None,
        markers=markers,
        progressive_mode="PROGRESSIVE_MODE" in flags
    )




# versions
from setuptools.command.build_ext import build_ext
import re
from pathlib import Path
import ctypes
import setuptools
import codecs
import os
__version__ = os.environ.get('VERSION_NEW', '0.10.9')
libjpeg_versions = {
    '6b': (None, 60),
    '7': (None, 70),
    '8': (None, 80),
    '8a': (None, 80),
    '8b': (None, 80),
    '8c': (None, 80),
    '8d': (None, 80),
    '9': (None, 90),
    '9a': (None, 90),
    '9b': (None, 90),
    '9c': (None, 90),
    '9d': (None, 90),
    '9e': (None, 90),
    'turbo210': ('2.1.0', 210),
    'mozjpeg403': ('4.0.3', 403)
}

# requirements
try:
    with open('requirements.txt') as f:
        reqs = f.read().splitlines()
except:
    reqs = ['numpy']

with codecs.open("README.md", "r", encoding="UTF-8") as fh:
    long_description = fh.read()

# create version dependent extensions
cfiles = {}
hfiles = {}
cjpeglib = {}
for v in libjpeg_versions:
    is_moz = v[:3] == "moz"
    is_turbo = v[:5] == "turbo" or is_moz

    clib = f'jpeglib/cjpeglib/{v}'

    # create missing
    package_name = 'libjpeg'
    (Path(clib) / 'jconfig.h').touch()
    (Path(clib) / 'vjpeglib.h').touch()
    if is_turbo:
        package_name += '-turbo'
        (Path(clib) / 'jconfigint.h').touch()

    files = [
        f'{clib}/{f}' for f in os.listdir(clib) if re.fullmatch(f'.*\.(c|h)', f)]
    for excluded_module in ['jmemdos', 'jmemmac', 'jmemansi', 'ansi2knr', 'ckconfig', 'jmemname',  # platform dependents
                            'djpeg', 'cjpeg', 'rdjpgcom', 'wrjpgcom', 'cdjpeg', 'jpegtran',  # executables
                            'rdbmp', 'wrbmp', 'rdcolmap', 'rdppm', 'wrppm', 'rdtarga', 'wrtarga', 'rdrle', 'wrrle', 'rdgif', 'wrgif', 'rdswitch',  # others
                            'example',  # example
                            'cjpegalt', 'djpegalt',
                            # 'jerror',
                            # turbo
                            'jccolext', 'jdcolext', 'jdcol565', 'jstdhuff',
                            'jdmrg565', 'jdmrgext', "jcstest", "tjunittest", "tjbench",
                            'turbojpeg-jni', 'turbojpeg']:
        lim = -2 - len(excluded_module)
        files = [f for f in files if f[lim:-2] != excluded_module]
    #
    cfiles[v] = [f for f in files if f[-2:] == '.c']
    hfiles[v] = [f for f in files if f[-2:] == '.h']
    sources = ['jpeglib/cjpeglib/cjpeglib.c', *cfiles[v]]

    macros = [
        ("BITS_IN_JSAMPLE", 8),
        ("HAVE_STDLIB_H", 1),
        ("LIBVERSION", libjpeg_versions[v][1]),
        ("HAVE_PROTOTYPES", 1),
    ]
    if is_turbo:
        macros += [
            ("JPEG_LIB_VERSION", 80),  # 70),
            ("INLINE", "__inline__"),
            ("PACKAGE_NAME", f"\"{package_name}\""),
            ("BUILD", f"\"unknown\""),
            ("VERSION", f"\"{libjpeg_versions[v][0]}\""),
            ("SIZEOF_SIZE_T", int(ctypes.sizeof(ctypes.c_size_t))),
            ("THREAD_LOCAL", "__thread")
        ]
    if is_moz:
        macros += [
            ('C_ARITH_CODING_SUPPORTED', 1)
        ]

    cjpeglib[v] = setuptools.Extension(
        name=f"jpeglib/cjpeglib/cjpeglib_{v}",
        library_dirs=['./jpeglib/cjpeglib', f'./{clib}'],  # [f'./{clib}'],
        include_dirs=['./jpeglib/cjpeglib', f'./{clib}'],  # [f'./{clib}'],
        sources=sources,
        headers=hfiles[v],
        define_macros=macros,
        extra_compile_args=["-fPIC", "-g"],
        language="c",
    )


class custom_build_ext(build_ext):
    def build_extensions(self):
        #self.compiler.set_executable("compiler_so", "g++")
        #self.compiler.set_executable("compiler_cxx", "g++")
        #self.compiler.set_executable("linker_so", "g++")
        # 'add_include_dir', 'add_library', 'add_library_dir', 'add_link_object', 'add_runtime_library_dir',
        # 'announce', 'archiver', 'compile', 'compiler', 'compiler_cxx', 'compiler_so', 'compiler_type',
        # 'create_static_lib', 'debug_print', 'define_macro', 'detect_language', 'dry_run', 'dylib_lib_extension',
        # 'dylib_lib_format', 'exe_extension', 'executable_filename', 'executables', 'execute', 'find_library_file',
        # 'force', 'has_function', 'include_dirs', 'language_map', 'language_order', 'libraries', 'library_dir_option',
        # 'library_dirs', 'library_filename', 'library_option', 'link', 'link_executable', 'link_shared_lib',
        # 'link_shared_object', 'linker_exe', 'linker_so', 'macros', 'mkpath', 'move_file', 'obj_extension',
        # 'object_filenames', 'objects', 'output_dir', 'preprocess', 'preprocessor', 'ranlib', 'runtime_library_dir_option',
        # 'runtime_library_dirs', 'set_executable', 'set_executables', 'set_include_dirs', 'set_libraries', 'set_library_dirs',
        # 'set_link_objects', 'set_runtime_library_dirs', 'shared_lib_extension', 'shared_lib_format', 'shared_object_filename',
        # 'spawn', 'src_extensions', 'static_lib_extension', 'static_lib_format', 'undefine_macro', 'verbose', 'warn',
        # 'xcode_stub_lib_extension', 'xcode_stub_lib_format'
        #print("==========", self.compiler.library_dirs)
        build_ext.build_extensions(self)


setuptools.setup(
    name='jpeglib',
    version=__version__,
    author=u'Martin Beneš',
    author_email='martinbenes1996@gmail.com',
    description='Python envelope for the popular C library libjpeg for handling JPEG files.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    license='MPL',
    #test_suite = 'setup.test_suite',
    url='https://jpeglib.readthedocs.io/en/latest/',
    #download_url = 'https://github.com/martinbenes1996/jpeglib/archive/0.1.0.tar.gz',
    keywords=['jpeglib', 'jpeg', 'jpg', 'libjpeg', 'compression',
              'decompression', 'dct-coefficients', 'dct'],
    install_requires=reqs,
    package_dir={'': '.'},
    package_data={'': ['data/*']},
    include_package_data=True,
    ext_modules=[cjpeglib[v] for v in libjpeg_versions],
    cmdclass={"build_ext": custom_build_ext},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: C',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Topic :: Education',
        'Topic :: Multimedia',
        'Topic :: Multimedia :: Graphics :: Graphics Conversion',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Security',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities'
    ],
)




import numpy as np

from ._base import *

def _dct_ii(X:np.ndarray, u:int, v:int):
    """DCT coefficient at index u,v.
    
    :param X: input pixel data tensor 
    :type X: np.ndarray
    :param u: horizontal index of DCT coefficient
    :type u: int
    :param v: vertical index of DCT coefficient
    :type v: int
    :return: output DCT coefficient value
    :rtype: float
    
    :Examples:
    
    >>> X = np.array([[1,2],[3,1]])
    >>> _dct_ii(X,1,0) # -> 2.72e-16
    """
    N,M = X.shape
    Yuv = (
        np.sqrt(2/N) * np.sqrt(2/M) *
        Lambda(u) * Lambda(v) *
        sum(
            X[i,j] * np.cos(np.pi/N*(i+.5)*u) * np.cos(np.pi/N*(j+.5)*v)
        for i in range(N)
        for j in range(M))
    )
    return Yuv

def _dct_iii(Y, i, j):
    """Pixel value at index i,j.
    
    :param x: input DCT coefficient tensor
    :type x: np.ndarray
    :param i: horizontal index of the pixel
    :type i: int
    :param j: vertical index of the pixel
    :type j: int
    :return: output value
    :rtype: float
    """
    N,M = Y.shape
    Xij = (
        sqrt(2/N) * sqrt(2/M) *
        sum(
            Lambda(u) * Lambda(v) * Y[u,v] * np.cos(np.pi/N*(i+.5)*u) * np.cos(np.pi/N*(j+.5)*v)
        for u in range(N)
        for v in range(M))
    )
    return Xij

def DCT2D(X:np.ndarray):
    """Python implementation of 2D DCT transformation.
    
    :param X: pixel data in spatial domain
    :type X: np.ndarray
    :return: tensor of DCT coefficients
    :rtype: np.ndarray
    
    :Examples:
    
    >>> X = np.array([[1,2],[3,1]])
    >>> Y = jpeglib.dct.DCT2D(X)
    """
    return np.array([[_dct_ii(X, u,v) for v in range(X.shape[1])] for u in range(X.shape[0])])

def iDCT2D(Y:np.ndarray):
    """Python implementation of 2D inverse DCT transformation.
    
    :param Y: tensor of DCT coefficients
    :type Y: np.ndarray
    :return: pixel data in spatial domain
    :rtype: np.ndarray
    
    :Examples:
    
    >>> X = np.array([[1,2],[3,1]])
    >>> Y = jpeglib.dct.DCT2D(X)
    >>> jpeglib.dct.iDCT2d(Y) # -> reconstruction of X
    """
    return np.array([[_dct_iii(Y, i,j) for j in range(Y.shape[1])] for i in range(Y.shape[0])])

__all__ = ["DCT2D", "iDCT2D"]